@@ -41,6 +41,7 @@
 #include <gdk/gdkx.h>
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
+#include <X11/Xatom.h>
 #include "main.h"
 #include "gok.h"
 #include "gok-word-complete.h"
@@ -62,6 +63,7 @@
 #include <signal.h>
 #include "gok-bounds.h"
 
+#define USE_GCONF 1
 
 #define APP_STATIC_BUFF_SZ 30
 #define MAX_DELETABLE_KEYBOARDS 50
@@ -73,6 +75,13 @@
 #define GOK_ERROR_DISPLAY_SCAN_MAIN   5
 #define GOK_ERROR_ACTION_OPEN         6
 #define GOK_ERROR_FEEDBACK_OPEN       7
+#define GOK_ERROR_NO_XKB_EXTENSION    8
+#define GOK_ERROR_NOSTICKYKEYS		  9
+
+#define GOKMAINKEYBOARDNAME "main"
+
+#define KEYBOARD_ACCESSIBILITY_STICKY_KEYS_KEY "/desktop/gnome/accessibility/keyboard/stickykeys_enable"
+#define GCONF_ACCESSIBILITY_KEY "/desktop/gnome/interface/accessibility"
 
 /* pointer to the first keyboard in the list of keyboards */
 static GokKeyboard* m_pKeyboardFirst = NULL;
@@ -113,40 +122,124 @@ void gok_main_initialize_commandprediction (void);
 gboolean gok_main_display_scan_main(void);
 void gok_main_display_geometry_error (void);
 
+/* poptOption storage */
+static char* accessmethodname = NULL;
+static char* inputdevicename = NULL;
+static char* mainkeyboardname = NULL;
 static int bDisplayKeyboardEditor = 0;
 static char *geometry = NULL;
 static int geometry_bitmask;
 static int bLoginOption = 0;
 static int bRememberGeometryOption = 0;
 static int bDisplaySettingsDialog = 0;
+static int bExtras = 0;
 
 static struct poptOption options[] = {
-	{"editor", 'e', POPT_ARG_NONE, &bDisplayKeyboardEditor, 0,
-		N_("Start the GOK keyboard editor"), NULL},
-	{"geometry", '\0', POPT_ARG_STRING, &geometry, 0,
+	{
+		"access-method", 
+		'a',
+		POPT_ARG_STRING, 
+		&accessmethodname, 
+		0,
+		N_("Use the specified access method. NAME is a string and can be found in the various access method files (.xam) assigned to the \"name\" property of <gok:accessmethod> tag. Note this is not necessarily the same as the name of the .xam file."), 
+		N_("NAME")
+	},
+	{
+		"editor", 
+		'e', 
+		POPT_ARG_NONE, 
+		&bDisplayKeyboardEditor, 
+		0,
+		N_("Start the GOK keyboard editor"), 
+		NULL},
+	{
+		"extras", 
+		'\0', 
+		POPT_ARG_NONE, 
+		&bExtras, 
+		0,
+		N_("Use special, but possibly unstable, gok stuff"),
+		NULL},
+	{
+		"geometry", 
+		'\0', 
+		POPT_ARG_STRING, 
+		&geometry, 
+		0,
 		N_("Whenever --geometry is not used gok remembers its position between invocations and starts in the position that it had when it was last shutdown.  When --geometry is used gok positions itself within the rectangular area of screen described by the given X11 geometry specification.  When --geometry is used gok does not remember its position when it shuts down.  This behaviour can be changed with the --remembergeometry flag which forces gok to remember its position when shutdown even when it was started with --geometry."),
-		N_("GEOMETRY")},
-	{"login", 'l', POPT_ARG_NONE, &bLoginOption, 0,
+		N_("GEOMETRY")
+	},
+	{
+		/* TODO: NOT USED */
+		"input-device", 
+		'i', 
+		POPT_ARG_STRING, 
+		&inputdevicename, 
+		0,
+		N_("Use the specified input device"), 
+		N_("DEVICENAME")
+	},
+	{
+		"keyboard", 
+		'k', 
+		POPT_ARG_STRING, 
+		&mainkeyboardname, 
+		0,
+		N_("Start GOK with the specified keyboard."), 
+		N_("KEYBOARDNAME")
+	},
+	{
+		/* TODO: NOT USED */
+		"login", 
+		'l', 
+		POPT_ARG_NONE, 
+		&bLoginOption, 
+		0,
 		N_("GOK will be used to login"),
-		NULL},
-	{"remembergeometry", '\0', POPT_ARG_NONE,
-		&bRememberGeometryOption, 0,
+		NULL
+	},
+	{
+		"remembergeometry", 
+		'\0', 
+		POPT_ARG_NONE,
+		&bRememberGeometryOption, 
+		0,
 		N_("Can be used with --geometry.  Forces GOK to remember its position when shutdown even when it was started with --geometry.  Please see the discussion under the --geometry flag for more information."),
-		NULL},
-	{"settings", 's', POPT_ARG_NONE, &bDisplaySettingsDialog, 0,
+		NULL
+	},
+	{
+		"settings", 
+		's', 
+		POPT_ARG_NONE, 
+		&bDisplaySettingsDialog, 
+		0,
 		N_("Open the settings dialog box when GOK starts"),
-		NULL},
+		NULL
+	},
 
 	/* End the list */
-	{NULL, '\0', 0, NULL, 0}
+	{NULL, '\0', 0, NULL, 0, NULL, NULL}
 };
 
+static int segfaults = 0;
 /* private */
-static int gok_segv_handler(int sig)
+static int gok_sig_handler(int sig)
 {
-	fprintf (stderr, "gok: Segmentation fault\n");
-
-	gok_main_close();
+	switch (sig) {
+	case SIGSEGV:
+		segfaults++;
+		fprintf (stderr, "gok: Segmentation fault\n");
+		break;
+	case SIGTERM:
+		fprintf (stderr, "gok: exiting (terminated)\n");
+		break;
+	case SIGINT:
+	default:
+		break;
+	}
+	
+	if (segfaults == 1)
+		gok_main_close();
 
 	_exit(1);
 	return -1;  /* this line gets rid of compiler warning */ 
@@ -181,12 +274,10 @@ gint main (gint argc, gchar *argv[])
 	}
 	
 	gok_spy_add_idle_handler ();
-
+	
 	/* start the main hook */
 	SPI_event_main();	
 
-	gok_spy_remove_idle_handler ();
-
 	/* cleanup */
 	gok_main_close();
 	 
@@ -217,6 +308,7 @@ gint gok_main_open(gint argc, gchar *argv[])
 		LIBGNOMEUI_MODULE, argc, argv,
 		GNOME_PARAM_POPT_TABLE, options,
 		GNOME_PROGRAM_STANDARD_PROPERTIES,
+	        GNOME_PARAM_APP_DATADIR, DATADIR"/gok", 
 		LIBGNOMEUI_PARAM_CRASH_DIALOG, FALSE,
 		NULL);
 
@@ -281,6 +373,13 @@ gint gok_main_open(gint argc, gchar *argv[])
 		SPI_exit();
 		return 1;
 	}
+	
+	/* check for xkb extension */
+	if (!gok_main_has_xkb_extension())
+	{
+		gok_main_display_error (_("XKB extension is required."));
+		return GOK_ERROR_NO_XKB_EXTENSION;
+	}		
 
 	/* read the keyboards */
 	gok_main_read_keyboards ();
@@ -320,7 +419,8 @@ gint gok_main_open(gint argc, gchar *argv[])
 	gok_sound_initialize();
 
 	/* create the main window */
-	m_pWindowMain = gok_main_create_window ();
+	m_pWindowMain = gok_main_create_window (
+		gok_data_get_dock_type () != GOK_DOCK_NONE);
 	if (m_pWindowMain == NULL)
 	{
 		gok_main_display_error (_("Can't create the main GOK window!"));
@@ -343,14 +443,39 @@ gint gok_main_open(gint argc, gchar *argv[])
 		return GOK_ERROR_SETTINGSDIALOG_OPEN;
 	}
 
+	if (mainkeyboardname == NULL)
+	{
+		mainkeyboardname = (gchar*) g_malloc (strlen (GOKMAINKEYBOARDNAME) + 1);
+		strcpy (mainkeyboardname, (gchar *) GOKMAINKEYBOARDNAME);
+	}
+	
+
+	gok_main_check_accessibility();
+	
+	if (!gok_main_check_sticky_keys (m_pWindowMain)) {
+		return GOK_ERROR_NOSTICKYKEYS;
+	}
+	
 	/* display the "main" keyboard */
 	if (gok_main_display_scan_main () == FALSE)
 	{
 		return GOK_ERROR_DISPLAY_SCAN_MAIN;
 	}
 
-	/* connect a segv handler */
-	signal(SIGSEGV, (void (*)(int))gok_segv_handler);
+gok_log ("check for currently active frame");
+	if (m_pForegroundWindowAccessible == NULL) {
+		/* check for currently active frame */
+		Accessible* accessible;
+		accessible = gok_spy_get_active_frame();
+		if (accessible)
+			gok_main_window_change_listener (accessible);
+	}
+gok_log ("finished check");
+	
+	/* connect signal handlers; don't trap SIGKILL */ 
+	signal(SIGSEGV, (void (*)(int))gok_sig_handler);
+	signal(SIGTERM, (void (*)(int))gok_sig_handler);
+	signal(SIGINT,  (void (*)(int))gok_sig_handler);
 
 	return 0;
 }
@@ -358,7 +483,7 @@ gint gok_main_open(gint argc, gchar *argv[])
 /* private */
 gboolean gok_main_display_scan_main()
 {
-	if (gok_main_display_scan ((GokKeyboard*)NULL, "main", KEYBOARD_TYPE_UNSPECIFIED, KEYBOARD_LAYOUT_NORMAL) == FALSE)
+	if (gok_main_display_scan ((GokKeyboard*)NULL, mainkeyboardname, KEYBOARD_TYPE_UNSPECIFIED, KEYBOARD_LAYOUT_NORMAL) == FALSE)
 	{
 		/* if no "main" keyboard then display first keyboard in list */
 		if (gok_main_display_scan ((GokKeyboard*)NULL, NULL, KEYBOARD_TYPE_UNSPECIFIED, KEYBOARD_LAYOUT_NORMAL) == FALSE)
@@ -467,6 +592,7 @@ Accessible* gok_main_get_foreground_window_accessible()
 	return m_pForegroundWindowAccessible;
 }
 
+
 /**
 * gok_main_window_change_listener
 * @pAccessible: Pointer to the foreground window's accessible interface.
@@ -479,33 +605,32 @@ void gok_main_window_change_listener (Accessible* pAccessible)
 	GokKeyboard* pKeyboard = NULL;
 
 	gok_log_enter();	
-	if (1) /* TODO check to make sure the new window is not the gok window! */
-	{
-		m_pForegroundWindowAccessible = pAccessible;
-		
-		/* we might currently have a dynamic keyboard so we need to branch to be safe */
-		gok_main_display_scan_previous_premade();
-		
-		/*if (*/gok_keyboard_validate_dynamic_keys (pAccessible);/* == TRUE)*/
-		{
-			/* rechunk the keyboard because keys have changed state */
-			gok_scanner_stop();
-			pKeyboard = gok_main_get_current_keyboard();
-			
-			gok_chunker_chunk (pKeyboard);
-			pKeyboard->bRequiresChunking = FALSE;
-		
-			/* highlight the first chunk */
-			gok_chunker_highlight_first_chunk();
-			
-			/* restart the scanning process */
-			gok_scanner_start();
-		}
 
-		/* hack to make sure gok is not occluded - seems to be unstable*/
-		gdk_window_raise (m_pWindowMain->window);
+	/* note: ref counting happens in gok-spy's listeners */
+	m_pForegroundWindowAccessible = pAccessible;
+	
+	/* we might currently have a dynamic keyboard so we need to branch to be safe */
+	gok_main_display_scan_previous_premade();
+	
+	/*if (*/gok_keyboard_validate_dynamic_keys (pAccessible);/* == TRUE)*/
+	{
+		/* rechunk the keyboard because keys have changed state */
+		gok_scanner_stop();
+		pKeyboard = gok_main_get_current_keyboard();
+		
+		gok_chunker_chunk (pKeyboard);
+		pKeyboard->bRequiresChunking = FALSE;
+	
+		/* highlight the first chunk */
+		gok_chunker_highlight_first_chunk();
+		
+		/* restart the scanning process */
+		gok_scanner_start();
 	}
 
+	/* hack to make sure gok is not occluded - seems to be unstable*/
+	gdk_window_raise (m_pWindowMain->window);
+
 	gok_log_leave();	
 }
 
@@ -523,6 +648,7 @@ void gok_main_motion_listener (gint n_axes, int *motion_data, long mods, long ti
         GtkWidget *window = gok_main_get_main_window ();
 	Display *display;
 	GdkWindow *root;
+	gint dx, dy;
 	if (window) 
 	{
 		display = GDK_WINDOW_XDISPLAY (window->window);
@@ -533,17 +659,13 @@ void gok_main_motion_listener (gint n_axes, int *motion_data, long mods, long ti
 			mods, timestamp);
 		gok_log_leave ();
 		gok_scanner_input_motion (motion_data, n_axes);
-		
-		/* 
-		 * TODO:
-		 * XWarpPointer is apparently too slow to do every time we get a 
-		 * motion event, so if we want to support "cursor latching" we need 
-		 * to coalesce these events.
-		 */
+
+		dx = motion_data[0] * gok_data_get_valuator_sensitivity ();
+		dy = motion_data[1] * gok_data_get_valuator_sensitivity ();
 		if (gok_data_get_drive_corepointer ()) {
 			XWarpPointer (display, None, 
 				      GDK_WINDOW_XWINDOW (root),
-				      0, 0, 0, 0, motion_data[0], motion_data[1]);
+				      0, 0, 0, 0, dx, dy);
 		}
 	}
 }
@@ -876,6 +998,36 @@ gboolean gok_main_display_scan_previous ()
 	return bReturnCode;
 }
 
+GokKeyboard *
+gok_main_keyboard_find_byname (const gchar *NameKeyboard)
+{
+	GokKeyboard *pKB = NULL;
+	
+	if (NameKeyboard == NULL)
+	{
+		/* if keyboard name is NULL then use first keyboard in list */
+		pKB = m_pKeyboardFirst;
+	}
+	else
+	{
+		/* find the keyboard in the list (according to its name) */
+		pKB = m_pKeyboardFirst;
+		while (pKB != NULL)
+		{
+			if (g_strcasecmp (pKB->Name, NameKeyboard) == 0)
+			{
+				/* found the keyboard */
+				break;
+			}
+			else
+			{
+				pKB = pKB->pKeyboardNext;
+			}
+		}
+	}
+	return pKB;
+}
+
 /**
 * gok_main_display_scan
 * @pKeyboard: If this is supplied it takes precedence over the name parameter (useful for dynamic keyboards with the same name)
@@ -909,28 +1061,7 @@ gboolean gok_main_display_scan (GokKeyboard* pKeyboard, gchar* NameKeyboard, Key
 	}
 	else
 	{
-		/* if keyboard name is NULL then use first keyboard in list */
-		if (NameKeyboard == NULL)
-		{
-			pKB = m_pKeyboardFirst;
-		}
-		else
-		{
-			/* find the keyboard in the list (according to its name) */
-			pKB = m_pKeyboardFirst;
-			while (pKB != NULL)
-			{
-				if (g_strcasecmp (pKB->Name, NameKeyboard) == 0)
-				{
-					/* found the keyboard */
-					break;
-				}
-				else
-				{
-					pKB = pKB->pKeyboardNext;
-				}
-			}
-		}
+		pKB = gok_main_keyboard_find_byname (NameKeyboard);
 	}
 
 	if (pKB == NULL)
@@ -981,6 +1112,10 @@ gboolean gok_main_ds (GokKeyboard* pKeyboard)
 	/* stop the scanning process */
 	gok_scanner_stop();
 
+	/* raise the window as a precaution */
+	if (m_pWindowMain && m_pWindowMain->window)
+		gdk_window_raise (m_pWindowMain->window);
+
 	/* create new keys for dynamic keyboards */
 	if (pKeyboard->bDynamicallyCreated == TRUE)
 	{
@@ -997,7 +1132,7 @@ gboolean gok_main_ds (GokKeyboard* pKeyboard)
 		gok_log_leave();
 		return FALSE;
 	}
-	
+		
 	/* display the keyboard */
 	gok_keyboard_display (pKeyboard, m_pCurrentKeyboard, m_pWindowMain, TRUE);
 	m_pCurrentKeyboard = pKeyboard;
@@ -1035,7 +1170,7 @@ gboolean gok_main_ds (GokKeyboard* pKeyboard)
 *
 * returns: A pointer to the window, NULL if it couldn't be created.
 */
-GtkWidget* gok_main_create_window ()
+GtkWidget* gok_main_create_window (gboolean is_dock)
 {
 	GtkWidget *window1;
 	GtkWidget *fixed1;
@@ -1052,7 +1187,7 @@ GtkWidget* gok_main_create_window ()
 				"allow_shrink", TRUE,
 				"border_width", 0,
 				NULL),
-			     "signal::realize", on_window1_realize, NULL,
+			     "signal::realize", on_window1_realize, is_dock,
 			     "signal::destroy", on_window1_destroy, NULL,
 			     NULL);
 
@@ -1239,7 +1374,7 @@ GokKeyboard* gok_main_read_keyboards_from_dir (const char *directory)
 			gok_log ("complete_path = %s", complete_path);
 			pKeyboard = gok_keyboard_read (complete_path);
 			g_free (complete_path);
-
+			
 			/* add this keyboard into the list of keyboards */
 			pKeyboard->pKeyboardPrevious = pKeyboardPrevious;
 			if (pKeyboardFirst == NULL)
@@ -1271,37 +1406,33 @@ void gok_main_initialize_access_methods ()
     GConfClient *gconf_client = NULL;
     GError *gconf_err = NULL;
     gchar *directory_name = NULL;
-    gboolean got_error = FALSE;
 
     gconf_client = gconf_client_get_default ();
     
     directory_name = gconf_client_get_string (gconf_client,
 					      GOK_GCONF_ACCESS_METHOD_DIRECTORY,
 					      &gconf_err);
-    if (directory_name == NULL)
-    {
-	gok_log_x ("Got NULL access method directory key from GConf");
+    if (directory_name == NULL) {
+		gok_log_x ("Got NULL access method directory key from GConf");
         gok_main_display_gconf_error ();
-	exit (2);
+		exit (2);
     }
-    else if (gconf_err != NULL)
-    {
-	gok_log_x ("Error getting access method directory key from GConf");
-	g_error_free (gconf_err);
-	got_error = TRUE;
+    else if (gconf_err != NULL) {
+		gok_log_x ("Error getting access method directory key from GConf");
+		g_error_free (gconf_err);
+ 		gok_main_display_error
+			(_("could not access method directory key from GConf!"));
+		exit(2);		
     }
-    else
-    {
-	got_error = !(gok_scanner_initialize(directory_name));
+    else if (!(gok_scanner_initialize(directory_name))) {
+ 		gok_main_display_error
+			(_("possibly unknown access method!"));
+		exit(2);
+			
     }
     
     g_free (directory_name);
     
-    if (got_error)
-    {
-	gok_main_display_error (_("gok_scanner_initialize failed!"));
-	exit (2);
-    }
 }
 
 /**
@@ -1385,6 +1516,7 @@ void gok_main_initialize_commandprediction ()
 /**
 * gok_main_resize_window
 * @pWindow: Pointer to the main window.
+* @pKeyboard: Pointer to the relevant (pending) keyboard (may be NULL).
 * @Width: Width of the new window.
 * @Height: Height of the new window.
 * 
@@ -1396,7 +1528,7 @@ void gok_main_initialize_commandprediction ()
 * window then m_countIgnoreConfigure will be set to 2 so that the window 
 * center is not changed.
 */
-void gok_main_resize_window (GtkWidget* pWindow, gint Width, gint Height)
+void gok_main_resize_window (GtkWidget* pWindow, GokKeyboard *pKeyboard, gint Width, gint Height)
 {
 	gint left;
 	gint top;
@@ -1425,9 +1557,35 @@ void gok_main_resize_window (GtkWidget* pWindow, gint Width, gint Height)
 		return;
 	}
 	
+	/* let the keyboard know that the resize was caused by us */
+	gok_keyboard_set_ignore_resize (TRUE);
+
+	
 	/* calculate the upper left corner of the window */
-	left = gok_data_get_keyboard_x() - (Width / 2);
-	top = gok_data_get_keyboard_y() - (Height / 2);
+	switch (gok_data_get_dock_type () ) { /* if we're a dock */
+	case GOK_DOCK_BOTTOM:
+	  top = gdk_screen_height ();
+	  left = 0;
+	  if ((pKeyboard->expand == GOK_EXPAND_ALWAYS) || 
+	      (gok_data_get_expand () && (pKeyboard->expand != GOK_EXPAND_NEVER))) {
+		  Width = gdk_screen_width ();
+		  gok_keyboard_set_ignore_resize (FALSE);
+	  }
+	  break;
+	case GOK_DOCK_TOP:
+	  top = 0;
+	  left = 0;
+	  if ((pKeyboard->expand == GOK_EXPAND_ALWAYS) || 
+	      (gok_data_get_expand () && (pKeyboard->expand != GOK_EXPAND_NEVER))) {
+		  Width = gdk_screen_width ();
+		  gok_keyboard_set_ignore_resize (FALSE);
+	  }
+	  break;
+	default:
+	  top = gok_data_get_keyboard_y() - (Height / 2);
+	  left = gok_data_get_keyboard_x() - (Width / 2);
+	  break;
+	}
 
 	/* add the frame dimension to the window dimension */
 	/* get the frame location */
@@ -1441,11 +1599,14 @@ void gok_main_resize_window (GtkWidget* pWindow, gint Width, gint Height)
 		frameX = (winX - rectFrame.x);
 		frameY = (winY - rectFrame.y);
 	}
-	else
+	else if (gok_data_get_dock_type () == GOK_DOCK_NONE)
 	{
 		frameX = 5;/* this is usually true */
 		frameY = 22;/* this is usually true */
 	}
+	else {
+	        frameX = frameY = 0;
+	}
 	left -= frameX;
 	top -= frameY;
 	
@@ -1474,9 +1635,6 @@ void gok_main_resize_window (GtkWidget* pWindow, gint Width, gint Height)
 		top = screenY - Height - frameY - (frameX * 2);
 	}
 	
-	/* let the keyboard know that the resize was caused by us */
-	gok_keyboard_set_ignore_resize (TRUE);
-	
 	/* ensure the window is within the given geometry */
 	if (m_bUseGivenGeometry == TRUE)
 	{
@@ -1493,11 +1651,83 @@ void gok_main_resize_window (GtkWidget* pWindow, gint Width, gint Height)
 	/* move/resize the main window */
 	gdk_window_move_resize ((GdkWindow*)pWindow->window, left, top, Width, Height);
 
+	if (gok_data_get_dock_type () != GOK_DOCK_NONE) {
+		/* if this is a "never expand" keyboard, unset struts! */
+		if (pKeyboard && (pKeyboard->expand == GOK_EXPAND_NEVER))
+			gok_main_update_struts (Width, 1);
+		else
+			gok_main_update_struts (Width, Height);
+	}
+
 	/* store these values */
 	m_OurResizeWidth = Width;
 	m_OurResizeHeight = Height;
 }
 
+void
+gok_main_set_wm_dock (gboolean is_dock)
+{
+  Atom atom_type[1], atom_window_type;
+  GtkWidget *widget = gok_main_get_main_window ();
+
+  if (widget) gtk_widget_hide (widget);
+  gdk_error_trap_push ();
+  atom_window_type = gdk_x11_get_xatom_by_name ("_NET_WM_WINDOW_TYPE");
+  if (is_dock) 
+	  atom_type[0] = gdk_x11_get_xatom_by_name ("_NET_WM_WINDOW_TYPE_DOCK");
+  else
+	  atom_type[0] = gdk_x11_get_xatom_by_name ("_NET_WM_WINDOW_TYPE_NORMAL");
+
+  XChangeProperty (GDK_WINDOW_XDISPLAY (widget->window), 
+		   GDK_WINDOW_XWINDOW (widget->window), 
+		   atom_window_type,
+		   XA_ATOM, 32, PropModeReplace,
+		   (guchar *) &atom_type, 1);
+  gdk_error_trap_pop ();
+  if (widget) gtk_widget_show (widget);
+
+  if (m_pCurrentKeyboard) {
+	  gok_keyboard_display (m_pCurrentKeyboard, m_pCurrentKeyboard,
+				gok_main_get_main_window(), TRUE); 
+          /* this should move us to appropriate screen edge */
+  }
+
+}
+
+/*
+ * TODO: GtkDoc this!
+ */
+void
+gok_main_update_struts (gint width, gint height)
+{
+	GtkWidget *widget = gok_main_get_main_window ();
+	Atom atom_strut = gdk_x11_get_xatom_by_name ("_NET_WM_STRUT");
+	guint32 struts[4];
+
+	switch (gok_data_get_dock_type ()) {
+	case GOK_DOCK_TOP:
+		struts[0] = struts[2] = struts[3] = 0;
+		struts[1] = height;
+		break;
+	case GOK_DOCK_BOTTOM:
+		struts[0] = struts[1] = struts[2]= 0;
+		struts[3] = height;
+		break;
+	default:
+		return;
+		break;
+	}
+	gdk_error_trap_push ();
+	/* this means that we are responsible for placing ourselves appropriately on screen */
+	if (widget && widget->window)
+		XChangeProperty (GDK_WINDOW_XDISPLAY (widget->window), 
+				 GDK_WINDOW_XWINDOW (widget->window), 
+				 atom_strut,
+				 XA_CARDINAL, 32, PropModeReplace,
+				 (guchar *) &struts, 4);
+	gdk_error_trap_pop ();
+}
+
 /**
 * gok_main_get_our_window_size
 * @pWidth: Pointer to a variable that will hold the window width.
@@ -1564,6 +1794,8 @@ void gok_main_close()
 	gok_spy_deregister_windowchangelistener ((void *)gok_main_window_change_listener);
 	/* the call to deregister_mousebuttonlistener is in on_window1_destroy */
 	
+	gok_spy_stop();
+	
 	gok_log("BEFORE DELETING KEYBOARDS");
 	gok_log("Keyboards news - deletes: [%d]",gok_keyboard_get_keyboards());
 	gok_log("Stack pushes - pops:      [%d]",gok_branchbackstack_pushes_minus_pops());
@@ -1598,6 +1830,8 @@ void gok_main_close()
 	gok_control_button_callback_close();
 	gok_wordcomplete_close();
 	gok_predictor_close(m_pCommandPredictor);
+	
+	_exit(0);
 }
 
 /**
@@ -1704,3 +1938,217 @@ gok_main_display (void)
 		return GDK_WINDOW_XDISPLAY (gok_main_get_main_window ()->window);
 	return NULL;
 }
+
+
+
+/**
+* gok_main_get_access_method_override
+*
+* Call this to see if an access method was passed in from the command line
+*
+* returns: name of access method or NULL
+*/
+gchar* 
+gok_main_get_access_method_override (void)
+{
+	return accessmethodname;
+}
+
+/**
+* gok_main_get_geometry
+*
+* Call this to see if the extras argument was passed in from the command line
+*
+* returns: gboolean
+*/
+gboolean 
+gok_main_get_extras (void)
+{
+	return bExtras;
+}
+
+/**
+* gok_main_get_login
+*
+* Call this to see if the extras argument was passed in from the command line
+*
+* returns: gboolean
+*/
+gboolean 
+gok_main_get_login (void)
+{
+	return bLoginOption;
+}
+
+
+/**
+* gok_main_has_xkb_extension
+*
+* Call this to see if xkb is enabled.
+* 
+* returns: gboolean
+*/
+gboolean
+gok_main_has_xkb_extension ()
+{
+	/* TODO/revisit - is this a good check? */
+	if (gok_keyboard_get_xkb_desc() != NULL)
+	{
+		return TRUE;
+	}
+	return FALSE;
+}
+
+gboolean
+gok_main_check_sticky_keys (GtkWidget *widget)
+{
+	int op_rtn, event_rtn, error_rtn;
+	Display *display = GDK_WINDOW_XDISPLAY (widget->window);
+	GtkWidget *dialog;
+	gboolean xkb_ok = FALSE;
+
+	g_assert (display);
+	if (XkbQueryExtension (display,
+			       &op_rtn, &event_rtn, &error_rtn, NULL, NULL)) {
+		XkbDescRec *xkb;
+		xkb = XkbGetMap (display, XkbAllComponentsMask, XkbUseCoreKbd);
+		if (xkb) {
+			XkbGetControls (display,
+					XkbControlsEnabledMask,
+					xkb);
+			xkb_ok = TRUE;
+		} 
+		if (!xkb || !(xkb->ctrls->enabled_ctrls & XkbStickyKeysMask)) {
+			/* Turn on sticky keys and warn user! */
+#if USE_GCONF
+			if (1) {
+				GConfClient *gconf_client = gconf_client_get_default ();
+				gok_gconf_set_bool (gconf_client, KEYBOARD_ACCESSIBILITY_STICKY_KEYS_KEY, 
+						    TRUE); /* turn on gconf key */
+#else
+			if (XkbChangeEnabledControls (display,
+						      XkbUseCoreKbd,
+						      XkbStickyKeysMask,
+						      XkbStickyKeysMask)) {
+#endif
+				xkb_ok = TRUE;
+				dialog = gtk_message_dialog_new (GTK_WINDOW (widget),
+								 GTK_DIALOG_DESTROY_WITH_PARENT,
+								 GTK_MESSAGE_INFO,
+								 GTK_BUTTONS_OK,
+								 "GOK has enabled Sticky Keys, which it requires.\n");
+				
+				/* Destroy the dialog when the user responds to it (e.g. clicks a button) */
+				g_signal_connect_swapped (GTK_OBJECT (dialog), "response",
+							  G_CALLBACK (gtk_widget_destroy),
+							  GTK_OBJECT (dialog));
+				gtk_widget_show_all (dialog);
+			}
+			else
+				xkb_ok = FALSE;
+		}
+		XkbFreeKeyboard (xkb, 0, True);
+	}
+	if (xkb_ok == FALSE) {
+                /* post an error dialog */
+		dialog = gtk_message_dialog_new (GTK_WINDOW (widget),
+						 GTK_DIALOG_DESTROY_WITH_PARENT,
+						 GTK_MESSAGE_ERROR,
+						 GTK_BUTTONS_CLOSE,
+						 "GOK cannot run because XKB display extension is missing.\n");		
+		g_signal_connect_swapped (GTK_OBJECT (dialog), "response",
+					  G_CALLBACK (gtk_widget_destroy),
+					  GTK_OBJECT (dialog));
+		gtk_widget_show_all (dialog);
+		return FALSE;
+	}
+	return TRUE;
+}
+
+
+
+static GtkWidget* acd = NULL; /* accessibility check dialog */
+
+void check_accessibility_cb ( GObject* o, gpointer* data )
+{
+	gtk_widget_destroy(acd);	
+	
+	if (strcmp((gchar*)data,"logout") == 0) {
+		gok_gconf_set_bool ( gok_data_get_gconf_client(),
+			GCONF_ACCESSIBILITY_KEY, TRUE );
+			gok_main_close();  /* FIXME: need to logout */
+			_exit (0);
+	}
+	else if (strcmp((gchar*)data,"quit") == 0) {
+		gok_main_close();
+		_exit (0);
+	}
+	else if (strcmp((gchar*)data,"continue") == 0) {
+		/* maybe change gok somehow to show user weakened status */
+	}
+
+}
+
+void
+gok_main_check_accessibility ()
+{
+    gboolean     accessibility_on;
+    GConfClient* client;
+	
+	client = gok_data_get_gconf_client();
+
+    /* check if accessibility flag is TRUE */
+	gok_gconf_get_bool ( client, GCONF_ACCESSIBILITY_KEY, &accessibility_on );
+        
+    if (!accessibility_on) {
+		GtkWidget*	button;
+		gint         response_id;
+		gboolean 	returnCode = TRUE;
+		
+		acd = gtk_message_dialog_new (
+			GTK_WINDOW(m_pWindowMain), 
+			GTK_DIALOG_DESTROY_WITH_PARENT,
+			GTK_MESSAGE_WARNING,
+			GTK_BUTTONS_NONE,			
+			_("Assistive Technology Support Is Not Enabled." 
+		"\n\n" 
+		"You can start GOK without enabling support for assistive technologies. "
+		"However, some of the features of the application might not be available." 
+		"\n\n"
+		"To enable support for assistive technologies "
+		"and log in to a new session with the change enabled, "
+		"click "
+		"Enable and Log Out" "."
+		"\n\n"
+		"To continue using GOK, " 
+		"click "
+		"Continue" "."
+		"\n\n"
+		"To quit GOK, " 
+		"click " 
+		"Close" "." 
+		"\n\n"));
+		   
+		button = gtk_button_new_with_label(_("Enable and Log Out"));
+		gtk_container_add (GTK_CONTAINER (GTK_DIALOG(acd)->action_area),
+						  button);
+		g_signal_connect(G_OBJECT (button), "clicked", 
+			G_CALLBACK (check_accessibility_cb), (gpointer) "logout");
+			
+		button = gtk_button_new_from_stock(GTK_STOCK_CLOSE);
+		gtk_container_add (GTK_CONTAINER (GTK_DIALOG(acd)->action_area),
+						  button);
+		g_signal_connect(G_OBJECT (button), "clicked", 
+			G_CALLBACK (check_accessibility_cb), (gpointer) "quit");
+
+		button = gtk_button_new_with_label(_("Continue"));
+		gtk_container_add (GTK_CONTAINER (GTK_DIALOG(acd)->action_area),
+						  button);
+		g_signal_connect(G_OBJECT (button), "clicked", 
+			G_CALLBACK (check_accessibility_cb), (gpointer) "continue");
+						  
+
+		gtk_widget_show_all (acd);
+		
+	}
+}
